@@ -24,13 +24,62 @@ class ManuscriptEditor:
         self.title = get_yaml_field(metadata_file, "title")
         self.keywords = get_yaml_field(metadata_file, "keywords")
 
+    @staticmethod
+    def prepare_paragraph(paragraph: list[str]) -> str:
+        """
+        It takes a list of sentences that are part of a paragraph and joins them
+        into a single string. The paragraph might have Equations, which are
+        between "$$" and have an identifier between curly brackets. There are
+        two kinds of paragraphs:
+        1) "Simple paragraphs" are a set of sentences with no Equations;
+        2) "Equation paragraphs" are a set of sentences where at least one
+        sentence has an Equation. When joining sentences with Equations, a
+        newline is added before and after the Equation.
+        """
+        paragraph_text = ""
+
+        paragraph = iter(paragraph)
+        for sentence in paragraph:
+            sentence = sentence.strip()
+
+            if sentence == "":
+                paragraph_text += "\n"
+            elif sentence.startswith("$$"):
+                # this is an equation
+                equation_sentences = [sentence]
+
+                sentence = next(paragraph, None)
+                while sentence is not None and not sentence.startswith("$$"):
+                    equation_sentences.append(sentence)
+                    sentence = next(paragraph, None)
+
+                if sentence is not None:
+                    equation_sentences.append(sentence)
+
+                paragraph_text += "\n".join(equation_sentences) + "\n"
+            else:
+                simple_sentences = [sentence]
+
+                sentence = next(paragraph, None)
+                while sentence is not None and sentence != "":
+                    simple_sentences.append(sentence)
+                    sentence = next(paragraph, None)
+
+                suffix = "\n"
+                if sentence == "":
+                    suffix += "\n"
+
+                paragraph_text += " ".join(simple_sentences) + suffix
+
+        return paragraph_text
+
     @staticmethod
     def revise_and_write_paragraph(
         paragraph: list[str],
         section_name: str,
         revision_model: ManuscriptRevisionModel,
         outfile=None,
-    ) -> None | tuple[str]:
+    ) -> None | tuple[str, str]:
         """
         Revises and writes a paragraph to the output file.
 
@@ -45,7 +94,19 @@ class ManuscriptEditor:
             the submitted paragraph and the revised paragraph.
         """
         # Process the paragraph and revise it with model
-        paragraph_text = " ".join(paragraph)
+        paragraph_text = ManuscriptEditor.prepare_paragraph(paragraph)
+
+        # revise paragraph only if it has all these properties: 1) it has at
+        # least two sentences, 2) it has in total at least 60 words
+        if not (len(paragraph) > 2 and len(paragraph_text.split()) > 60):
+            paragraph_text = ManuscriptEditor.convert_sentence_ends_to_newlines(
+                paragraph_text
+            )
+            if outfile is not None:
+                outfile.write(paragraph_text)
+                return
+            else:
+                return paragraph_text, paragraph_text
 
         error_message = None
         try:
@@ -169,6 +230,7 @@ ERROR: the paragraph below could not be revised with the AI model due to the fol
             paragraph = []
 
             current_table_paragraph = False
+            last_sentence_ends_with_alphanum_or_colon = False
 
             for line in infile:
                 if line.startswith("<!--"):
@@ -221,21 +283,67 @@ ERROR: the paragraph below could not be revised with the AI model due to the fol
 
                 # If the line is blank, it indicates the end of a paragraph
                 elif line.strip() == "":
+                    # if the last sentence added to the paragraph ends with an
+                    # alphanumeric character or a colon (i.e., does not end with a
+                    # period as normal sentences), then we keep adding all lines to
+                    # to the paragraph until we find a line with text that starts
+                    # with a capital letter and is preceded by an empty line.
+                    if last_sentence_ends_with_alphanum_or_colon:
+                        while line is not None and not (
+                            prev_line.strip() == ""
+                            and (
+                                (line[0].isalnum() and line[0].isupper())
+                                or line.startswith("#")
+                            )
+                        ):
+                            paragraph.append(line.strip())
+                            prev_line = line
+                            line = next(infile, None)
+
+                        last_sentence_ends_with_alphanum_or_colon = False
+
+                        # remove all trailing empty sentences in the paragraph,
+                        # and for each write a new line to the output file
+                        prev_line = ""
+                        while len(paragraph) > 0 and paragraph[-1] == "":
+                            paragraph.pop()
+                            prev_line += "\n"
+
                     # revise and write paragraph to output file
                     self.revise_and_write_paragraph(
                         paragraph, section_name, revision_model, outfile
                     )
 
-                    # and also the current line, which is the end of the
-                    # paragraph
-                    outfile.write(line)
-
                     # clear the paragraph list
-                    paragraph = []
+                    if line.strip() == "":
+                        outfile.write(line)
+                        paragraph = []
+                    else:
+                        outfile.write(prev_line)
+
+                        if line.startswith("#"):
+                            outfile.write(line)
+                            paragraph = []
+                        else:
+                            paragraph = [line.strip()]
 
                 # Otherwise, add the line to the paragraph list
                 else:
-                    paragraph.append(line.strip())
+                    line_strip = line.strip()
+
+                    # check if line ends with a colon, a letter or a number:
+                    last_sentence_ends_with_alphanum_or_colon = (
+                        not self.line_is_not_part_of_paragraph(line_strip)
+                        and (
+                            line_strip.endswith(":")
+                            or line_strip[-1].isalpha()
+                            or line_strip[-1].isdigit()
+                        )
+                    )
+
+                    paragraph.append(line_strip)
+
+                prev_line = line
 
             # If there's any remaining paragraph, process and write it to the
             # output file
